import itertools
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence

from musictool import config


def iter_scales(kind, start=None):
    scales = getattr(config, kind)
    it = itertools.cycle(scales)
    if start:
        it = itertools.dropwhile(lambda x: x != start, it)
    it = itertools.islice(it, len(scales))
    return list(it)


def unique(iterable, key=lambda x: x):
    seen = set()
    for item in iterable:
        ki = key(item)
        if ki in seen:
            continue
        seen.add(ki)
        yield item


def sequence_builder(
    n: int,
    options: Iterable | Sequence[Iterable],
    options_separated: bool = False,
    curr_prev_constraint: Callable | None = None,
    candidate_constraint: Callable | None = None,
    i_constraints: dict[int, Callable] = dict(),
    unique_key: Callable | None = None,
    loop: bool = True,
    prefix: Sequence = tuple(),
) -> Sequence:
    """build a sequence of elements from options
    Parameters
    ----------
    n : int
        len of returned sequence
    options: Iterable
        on each build step use one of the `options`
    options_separated : bool
        if True options should be a Sequence of separate options for each step
    curr_prev_constraint
    candidate_constraint
    i_constraints
    unique_key: Callable | None
        if not None, sequence elements should have unique keys
    loop : bool
        if True then also checks curr_prev_constraint(seq[-1], seq[0])
    prefix:

    Returns
    -------
    Sequence
    """
    seq = list(prefix) if prefix else list()

    ops = options[len(seq)] if options_separated else options

    if i_constraint := i_constraints.get(len(seq)):
        ops = filter(i_constraint, ops)

    if unique_key:
        prefix_keys = frozenset(unique_key(op) for op in prefix)
        ops = frozenset(op for op in ops if unique_key(op) not in prefix_keys)

    if len(seq) > 0 and curr_prev_constraint is not None:
        ops = (op for op in ops if curr_prev_constraint(seq[-1], op))

    for op in ops:
        candidate = seq + [op]
        if candidate_constraint is not None and not candidate_constraint(candidate):
            continue
        if len(candidate) == n:
            if curr_prev_constraint is not None and loop and not curr_prev_constraint(candidate[-1], candidate[0]):
                continue
            yield tuple(candidate)
        else:
            yield from sequence_builder(n, options, options_separated, curr_prev_constraint, candidate_constraint, i_constraints, unique_key, loop, prefix=candidate)



import functools
import itertools

from musictool.chord import SpecificChord


@functools.cache
def have_parallel_interval(a: SpecificChord, b: SpecificChord, interval: int) -> bool:
    '''
    parallel in same voices!
    if there'are eg fifth in 1st and fifth in 2nd chord but not from same voices
    - then it allowed (aint considered parallel) (test it)

    a1 - b1
    a0 - b0
    todo: what about fifths + octave (eg C5 G6 -> F5 C6)
    '''
    voice_transitions = tuple(zip(a.notes_ascending, b.notes_ascending))
    for (a0, b0), (a1, b1) in itertools.combinations(voice_transitions, 2):
        if abs(a0 - a1) % 12 == interval == abs(b0 - b1) % 12:
            return True
    return False


@functools.cache
def have_hidden_parallel(a: SpecificChord, b: SpecificChord, interval: int) -> bool:
    """
    hidden/direct parallel/consecutive interval is when:
        1. outer voices (lower and higher) go in same direction (instead of oblique or contrary motion)
        2. they approach param:interval
    voice leading rules often forbid hidden fifths and octaves (param:interval = 7, 0) (explanation: 12 % 12 == 0 octave equal to unison)
    """
    a_low, a_high = a.notes_ascending[0], a.notes_ascending[-1]
    b_low, b_high = b.notes_ascending[0], b.notes_ascending[-1]

    is_same_direction = (a_low < b_low and a_high < b_high) or (a_low > b_low and a_high > b_high)
    if is_same_direction and (b_high - b_low) % 12 == interval:
        return True
    return False


@functools.cache
def have_voice_overlap(a: SpecificChord, b: SpecificChord) -> bool:
    n = len(b.notes_ascending)
    for i in range(n):
        upper = i < n - 1 and b.notes_ascending[i] > a.notes_ascending[i + 1]
        lower = i > 0 and b.notes_ascending[i] < a.notes_ascending[i - 1]
        if upper or lower:
            return True
    return False


@functools.cache
def have_large_leaps(a: SpecificChord, b: SpecificChord, interval: int) -> bool:
    return any(
        abs(an - bn) > interval
        for an, bn in zip(a.notes_ascending, b.notes_ascending)
    )


@functools.cache
def no_large_spacing(c: SpecificChord, max_interval=12):
    return all(c.notes_ascending[i] - c.notes_ascending[i - 1] <= max_interval for i in range(1, len(c.notes_ascending)))



from __future__ import annotations

import itertools
import typing

from musictool import config
from musictool.note import Note
from musictool.note import SpecificNote

if typing.TYPE_CHECKING:
    from musictool.scale import ComparedScales
    from musictool.scale import Scale


class Piano:
    def __init__(
        self,
        size: tuple[int, int] = config.piano_img_size,
        scale: Scale | ComparedScales | None = None,
        red_notes: frozenset[Note] = frozenset(),
        green_notes: frozenset[Note] = frozenset(),
        blue_notes: frozenset[Note] = frozenset(),
        notes_squares: dict[Note: str] | None = None,
    ):
        if notes_squares is None:
            notes_squares = dict()
        self.size = size
        ww = self.size[0] // 14  # white key width
        bw = int(ww * 0.6)  # black key width
        bh = int(self.size[1] * 0.6)  # black key height
        do = config.default_octave
        white_notes = tuple(SpecificNote(config.chromatic_notes[i], octave) for octave, i in itertools.product((do, do + 1), (0, 2, 4, 5, 7, 9, 11)))
        black_notes = tuple(SpecificNote(config.chromatic_notes[i], octave) for octave, i in itertools.product((do, do + 1), (1, 3, 6, 8, 10)))

        WHITE_COLOR = (170,) * 3
        BLACK_COLOR = (80,) * 3
        RED_COLOR = 255, 0, 0
        GREEN_COLOR = 0, 255, 0
        BLUE_COLOR = 0, 0, 255

        for note in white_notes: note.color = WHITE_COLOR
        for note in black_notes: note.color = BLACK_COLOR

        SMALL_RECT_HEIGHT = 5
        SMALL_SQUARE_SIZE = 12

        self.rects = []

        # todo: merge white keys and black keys logic into single loop to deduplicate logic

        # white keys
        for note, x in zip(white_notes, range(0, self.size[0], ww)):
            color = note.color
            if scale is not None and note.abstract in scale.notes:
                color = scale.note_colors[note.abstract]
            self.rects.append(f"""<rect x='{x}' y='0' width='{x + ww}' height='{self.size[1]}' style='fill:rgb{color};stroke-width:1;stroke:rgb{BLACK_COLOR}' onclick="play_note('{note.abstract.name}', '{note.octave}')"/>""")

            if note.abstract in red_notes: self.rects.append(f"""<rect x='{x}' y='0' width='{x + ww}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{RED_COLOR};'/>""")
            if note.abstract in green_notes: self.rects.append(f"""<rect x='{x}' y='0' width='{x + ww}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{GREEN_COLOR};'/>""")
            if note.abstract in blue_notes: self.rects.append(f"""<rect x='{x}' y='0' width='{x + ww}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{BLUE_COLOR};'/>""")

            if (fill_border_color := notes_squares.get(note.abstract)):
                fill, border, text_color, str_chord = fill_border_color
                self.rects.append(f"""
                    <g onclick=play_chord('{str_chord}')>
                        <rect x='{(x + x + ww) // 2 - SMALL_SQUARE_SIZE // 2}' y='{self.size[1] - SMALL_SQUARE_SIZE - 5}' width='{SMALL_SQUARE_SIZE}' height='{SMALL_SQUARE_SIZE}' style='fill:rgb{fill};stroke-width:1;stroke:rgb{border}'/>
                        <text x='{(x + x + ww) // 2 - SMALL_SQUARE_SIZE // 2}' y='{self.size[1] - SMALL_SQUARE_SIZE - 5 + SMALL_SQUARE_SIZE}' font-family="Menlo" font-size='15' style='fill:rgb{text_color}'>{note.name}</text>
                    </g>
                """)

        # black notes
        it = (x for i, x in enumerate(range(0 + ww, self.size[0], ww)) if i not in {2, 6, 9, 13})
        for note, x in zip(black_notes, it):
            color = note.color
            if scale is not None and note.abstract in scale.notes:
                color = scale.note_colors[note.abstract]
            self.rects.append(f"""<rect x='{x - bw // 2}', y='0' width='{bw}' height='{bh}' style='fill:rgb{color};stroke-width:1;stroke:rgb{BLACK_COLOR}' onclick="play_note('{note.name}', '{note.octave}')"/>""")

            if note.abstract in red_notes: self.rects.append(f"""<rect x='{x - bw // 2}' y='0' width='{bw}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{RED_COLOR};'/>""")
            if note.abstract in green_notes: self.rects.append(f"""<rect x='{x - bw // 2}' y='0' width='{bw}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{GREEN_COLOR};'/>""")
            if note.abstract in blue_notes: self.rects.append(f"""<rect x='{x - bw // 2}' y='0' width='{bw}' height='{SMALL_RECT_HEIGHT}' style='fill:rgb{BLUE_COLOR};'/>""")

            if (fill_border_color := notes_squares.get(note.abstract)):
                fill, border, text_color, str_chord = fill_border_color
                self.rects.append(f"""
                    <g onclick=play_chord('{str_chord}')>
                        <rect x='{x - SMALL_SQUARE_SIZE // 2}' y='{bh - 15}' width='{SMALL_SQUARE_SIZE}' height='{SMALL_SQUARE_SIZE}' style='fill:rgb{fill};stroke-width:1;stroke:rgb{border}'/>
                        <text x='{x - SMALL_SQUARE_SIZE // 2}' y='{bh - 15 + SMALL_SQUARE_SIZE}' font-family="Menlo" font-size='15' style='fill:rgb{text_color}'>{note.name}</text>
                    </g>
                """)

        # border around whole svg
        self.rects.append(f"<rect x='0', y='0' width='{self.size[0] - 1}' height='{self.size[1] - 1}' style='fill:none;stroke-width:1;stroke:rgb{BLACK_COLOR}'/>")

    def add_rect(self, x, y, w, h, fill, border_color):
        pass

    def __repr__(self):
        return 'Piano'

    def _repr_svg_(self):
        rects = '\n'.join(self.rects)
        return f'''
        <svg width='{self.size[0]}' height='{self.size[1]}'>
        {rects}
        </svg>
        '''



import itertools
from collections.abc import Iterable

from musictool import config
from musictool.note import AnyNote
from musictool.note import Note
from musictool.note import SpecificNote


def iterate(
    start_note: AnyNote = config.chromatic_notes[0],
    take_n: int | None = None,
) -> Iterable[Note | SpecificNote]:
    names = itertools.cycle(config.chromatic_notes)

    if isinstance(start_note, SpecificNote):
        octaves = itertools.chain.from_iterable(
            itertools.repeat(octave, 12)
            for octave in itertools.count(start=start_note.octave)
        )
        notes = (SpecificNote(name, octave) for name, octave in zip(names, octaves))
    else:
        if isinstance(start_note, str):
            start_note = Note(start_note)
        notes = (Note(name) for name in names)

    notes = itertools.dropwhile(lambda note: note.name != start_note.name, notes)

    if take_n is not None:
        notes = itertools.islice(notes, take_n)

    yield from notes


def sort_notes(it: Iterable[AnyNote], start: AnyNote | None = None) -> tuple:
    first = next(iter(it))
    if isinstance(first, str): out = sorted(it, key=config.note_i.__getitem__)
    elif type(first) is Note: out = sorted(it, key=lambda note: note.i)
    elif type(first) is SpecificNote: out = sorted(it, key=lambda note: note.absolute_i)
    else: raise TypeError

    out = tuple(out)

    if start is None:
        return out

    i = out.index(start)
    return out[i:] + out[:i]



import random


def hex_to_rgb(color):
    return tuple(int(color[i:i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(color):
    return '#{:02x}{:02x}{:02x}'.format(*color)


def rgba_to_rgb(rgb_background, rgba_color):
    '''https://stackoverflow.com/a/21576659/4204843'''

    alpha = rgba_color[3]

    return (
        int((1 - alpha) * rgb_background[0] + alpha * rgba_color[0]),
        int((1 - alpha) * rgb_background[1] + alpha * rgba_color[1]),
        int((1 - alpha) * rgb_background[2] + alpha * rgba_color[2]),
    )


def random_rgb():
    return random.randrange(255), random.randrange(255), random.randrange(255)


def random_rgba():
    return random.randrange(255), random.randrange(255), random.randrange(255), 255



import asyncio
import itertools
import random
from numbers import Number

from musictool import config
from musictool.note import Note
from musictool.note import SpecificNote
from musictool.noteset import NoteSet


class Chord(NoteSet):
    intervals_to_name = {
        # triads
        frozenset({4, 7}): 'major',
        frozenset({3, 7}): 'minor',
        frozenset({3, 6}): 'diminished',
        # 7th
        frozenset({4, 7, 11}): 'maj7',
        frozenset({4, 7, 10}): '7',
        frozenset({3, 7, 10}): 'min7',
        frozenset({3, 6, 10}): 'half-dim-7',
        frozenset({3, 6, 9}): 'dim-7',
        # 6th
        frozenset({4, 7, 9}): '6',
        frozenset({3, 7, 9}): 'm6',
        # etc
        frozenset({4, 8}): 'aug',
        frozenset({2, 7}): 'sus2',
        frozenset({5, 7}): 'sus4',

    }
    name_to_intervals = {v: k for k, v in intervals_to_name.items()}


class SpecificChord:
    def __init__(
        self,
        notes: frozenset[SpecificNote],
        root: Note | None = None,
    ):
        self.notes = notes
        self.root = root
        self.abstract = Chord(frozenset(note.abstract for note in notes), root)
        self.root_specific = frozenset(note for note in notes if note.abstract == root)

        self.notes_ascending = tuple(sorted(notes, key=lambda note: note.absolute_i))
        self.intervals = tuple(note - self.notes_ascending[0] for note in self.notes_ascending[1:])  # from lowest note
        self.key = self.notes, self.root
        self.str_chord = '_'.join(repr(note) for note in self.notes_ascending)

    @classmethod
    def random(cls, n_notes=None, octaves=None):
        if n_notes is None:
            n_notes = random.randint(2, 5)
        if octaves is None:
            octaves = 3, 4, 5
        notes_space = tuple(
            SpecificNote(note, octave)
            for note, octave in itertools.product(config.chromatic_notes, octaves)
        )
        notes = frozenset(random.sample(notes_space, n_notes))
        return cls(notes)

    @classmethod
    def from_str(cls, string: str):
        notes, _, root = string.partition('/')
        root = Note(root) if root else None
        notes = frozenset(SpecificNote.from_str(note) for note in notes.split('_'))
        return cls(notes, root)

    def notes_combinations(self, ids=False):
        if ids: yield from itertools.combinations(range(len(self.notes_ascending)), 2)
        else: yield from itertools.combinations(self.notes_ascending, 2)
        # for n, m in itertools.combinations(self.notes_ascending, 2):
        #     yield n, m

    def find_intervals(self, interval: int):
        return tuple((n, m) for n, m in self.notes_combinations() if abs(m - n) == interval)

    # def add_note(self, note: Note | SpecificNote, steps: int):
    #     notes = self.notes_ascending
    #     if type(note) is Note:
    #         return notes[(notes.index(note) + steps) % len(notes)]
    #     else:
    #         raise TypeError

    def __repr__(self):
        _ = self.str_chord
        if self.root is not None:
            _ = f'{_}/{self.root.name}'
        return _

    def __eq__(self, other): return self.key == other.key
    def __hash__(self): return hash(self.key)

    # def __sub__(self, other):
    #     """
    #     https://music.stackexchange.com/a/77630
    #     considering no voice crossing
    #     """
    #     return sum(note.absolute_i for note in self.notes) - sum(note.absolute_i for note in other.notes)

    def __sub__(left, right):
        return sum(abs(l.absolute_i - r.absolute_i) for l, r in zip(left.notes_ascending, right.notes_ascending))

    async def play(self, seconds: Number = 1, bass_octave: int | None = None) -> None:
        tasks = [note.play(seconds) for note in self.notes]
        if bass_octave:
            if self.root is None:
                raise ValueError('cannot play bass when root is None')
            tasks.append(SpecificNote(self.root, bass_octave).play(seconds))
        await asyncio.gather(*tasks)

    def to_midi(self, path=None, n_bars=1):
        import mido
        mid = mido.MidiFile(type=0, ticks_per_beat=96)
        track = mido.MidiTrack()
        track.append(mido.MetaMessage(type='track_name', name='test_name'))
        track.append(mido.MetaMessage(type='time_signature', numerator=4, denominator=4, clocks_per_click=36))
        track.append(mido.MetaMessage(type='time_signature', numerator=4, denominator=4, clocks_per_click=36))
        if self.root is not None:
            track.append(mido.MetaMessage(type='marker', text=self.root.name))

        stop_time = int(n_bars * mid.ticks_per_beat * 4)

        for note in self.notes_ascending:
            track.append(mido.Message('note_on', note=note.absolute_i, velocity=100, time=0))
        for i, note in enumerate(self.notes_ascending):
            track.append(mido.Message('note_off', note=note.absolute_i, velocity=100, time=stop_time if i == 0 else 0))

        mid.tracks.append(track)
        mid.meta = {'chord': self}
        if path is None:
            return mid
        mid.save(path)



def cprint(*args, color=None, **kwargs):
    colormap = dict(
        BLACK='30m',
        RED='31m',
        GREEN='32m',
        YELLOW='33m',
        BLUE='34m',
        MAGENTA='35m',
        CYAN='36m',
        WHITE='37m',
        UNDERLINE='4m',
    )
    if color is None:
        print(*args, **kwargs)
    else:
        print('\033[' + colormap[color], end='')
        print(*args, **kwargs)
        print('\033[0m', end='')


def ago(e):
    # e: pass timedelta between timestamps in 1579812524 format
    e *= 1000  # convert to 1579812524000 format
    t = round(e / 1000)
    n = round(t / 60)
    r = round(n / 60)
    o = round(r / 24)
    i = round(o / 30)
    a = round(i / 12)
    if e < 0: return 'just now'
    elif t < 10: return 'just now'
    elif t < 45: return str(t) + ' seconds ago'
    elif t < 90: return 'a minute ago'
    elif n < 45: return str(n) + ' minutes ago'
    elif n < 90: return 'an hour ago'
    elif r < 24: return str(r) + ' hours ago'
    elif r < 36: return 'a day ago'
    elif o < 30: return str(o) + ' days ago'
    elif o < 45: return 'a month ago'
    elif i < 12: return str(i) + ' months ago'
    elif i < 18: return 'a year ago'
    else: return str(a) + ' years ago'



import itertools
import random

import pipe21 as P

from musictool import chromatic
from musictool import config
from musictool.note import AnyNote
from musictool.note import Note
from musictool.note import SpecificNote

'''
NoteSet
    NotesWithRoot
        Scale
        Chord
'''


def bits_to_intervals(bits: str) -> frozenset:
    return (
        bits
        | P.Map(int)
        | P.Pipe(enumerate)
        | P.Pipe(lambda it: itertools.islice(it, 1, None))
        | P.FilterValues()
        | P.Keys()
        | P.Pipe(frozenset)
    )


def intervals_to_bits(intervals: frozenset) -> str:
    bits = ['0'] * 12
    bits[0] = '1'
    for i in intervals:
        bits[i] = '1'
    return ''.join(bits)


class NoteSet:
    intervals_to_name: dict = {}
    name_to_intervals: dict = {}

    def __init__(
        self,
        notes: frozenset[str | Note],
        root: str | Note | None = None,
    ):
        """
        an unordered set of notes
        root:
            root note is optional
            notes w/o root has no intervals
        """

        if len(notes) == 0:
            raise ValueError('notes should be not empty')

        if isinstance(next(iter(notes)), str):
            notes = frozenset(Note(note) for note in notes)

        if isinstance(root, str):
            root = Note(root)

        if root is not None and root not in notes:
            raise KeyError('root should be one of notes')

        self.notes = notes
        self.root = root
        self.key = self.notes, self.root
        self.notes_ascending = chromatic.sort_notes(self.notes, start=self.root)
        self.str_chord = ''.join(note.name for note in self.notes_ascending)

        if root is not None:
            if root not in notes:
                raise ValueError('root note should be one of the chord notes')

            self.intervals = frozenset(note - root for note in notes - {root})
            self.bits = intervals_to_bits(self.intervals)
            self.name = self.__class__.intervals_to_name.get(self.intervals)

    @property
    def rootless(self): return NoteSet(self.notes)

    @classmethod
    def from_name(cls, root: str | Note, name: str):
        if isinstance(root, str):
            root = Note(root)
        notes = frozenset(root + interval for interval in cls.name_to_intervals[name]) | {root}
        return cls(notes, root)

    @classmethod
    def from_intervals(cls, root: str | Note, intervals: frozenset):
        if isinstance(root, str):
            root = Note(root)
        return cls(frozenset(root + interval for interval in intervals) | {root}, root)

    @classmethod
    def random(cls, n_notes=None):
        if n_notes is None:
            n_notes = random.randint(2, 5)
        notes = frozenset(random.sample(config.chromatic_notes, n_notes))
        return cls(notes)

    @classmethod
    def from_str(cls, string: str):
        notes, _, root = string.partition('/')
        root = Note(root) if root else None
        notes = frozenset(Note(note) for note in notes)
        return cls(notes, root)

    def add_note(self, note: AnyNote, steps: int) -> Note | SpecificNote:
        notes = self.notes_ascending

        if type(note) is str:
            if len(note) == 1:
                note = Note(note)
            elif len(note) == 2:
                note = SpecificNote.from_str(note)
            else:
                raise ValueError('invalid note string representation')

        if type(note) is Note:
            return notes[(notes.index(note) + steps) % len(notes)]
        elif type(note) is SpecificNote:
            # raise NotImplementedError
            octave, i = divmod(notes.index(note.abstract) + steps, len(notes))
            return SpecificNote(notes[i], note.octave + octave)
        else:
            raise TypeError

    def __eq__(self, other): return self.key == other.key
    def __hash__(self): return hash(self.key)
    def __len__(self): return len(self.notes)
    def __contains__(self, item): return item in self.notes
    # def __str__(self): return ''.join(note.name for note in self.notes)

    # def to_piano_image(self, base64=False):
    #     return Piano(chord=self)._repr_svg_()

    def __repr__(self):
        _ = self.str_chord
        if self.root is not None:
            _ += f'/{self.root.name}'
        return _

    # async def play(self, seconds=1):
    #     await SpecificChord(
    #         notes=frozenset(SpecificNote(note) for note in self.notes),
    #         root=self.root,
    #     ).play(seconds)
    #     notes_to_play = self.specific_notes
    #
    #     if bass:
    #         notes_to_play = itertools.chain(notes_to_play, [Note(self.root.name, octave=self.root.octave + bass)])
    #
    #     tasks = tuple(note.play(seconds) for note in notes_to_play)
    #     await asyncio.gather(*tasks)

    # def _repr_html_(self):
    #     label = hasattr(self, 'label') and f"id={self.label!r}"or ''
    #     number = hasattr(self, 'number') and self.number or ''
    #
    #     return f'''
    #     <li class='card {self.name}' {label}>
    #     <a href='play_chord_{self.str_chord}'>
    #     <span class='card_header' ><h3>{number} {self.root} {self.name}</h3></span>
    #     <img src='{self.to_piano_image(base64=True)}' />
    #     </a>
    #     </li>
    #     '''

    # def __repr__(self):
    #     label = hasattr(self, 'label') and f"id={self.label!r}"or ''
    #     number = hasattr(self, 'number') and self.number or ''
    #
    #     return f'''
    #     <li class='card {self.name}' {label} onclick=play_chord('{str(self)}')>
    #     <span class='card_header' ><h3>{number} {self.root} {self.name}</h3></span>
    #     <img src='{self.to_piano_image(base64=True)}' />
    #     </li>
    #     '''

# class SpecificNotes: pass


def note_range(
    start: SpecificNote,
    stop: SpecificNote,
    noteset: NoteSet | None = None,
) -> tuple[SpecificNote]:
    """returned range is including both ends (start, stop)"""
    if noteset is None:
        noteset = NoteSet(config.chromatic_notes)
    out = []
    note = start
    while True:
        out.append(note)
        if note == stop:
            break
        note = noteset.add_note(note, 1)
    return tuple(out)



import collections
import functools
import itertools
import operator
import random
from collections.abc import Iterable

import pipe21 as P

from musictool import config
from musictool.chord import Chord
from musictool.chord import SpecificChord
from musictool.note import SpecificNote
from musictool.scale import Scale
from musictool.scale import parallel
from musictool.scale import relative
from musictool.util.iteration import sequence_builder
from musictool.util.iteration import unique
from musictool.voice_leading import checks


def all_triads(octaves=(4, 5, 6)):
    all_notes = tuple(
        SpecificNote(note, octave)
        for octave, note in itertools.product(octaves, config.chromatic_notes)
    )
    n3 = tuple(itertools.combinations(all_notes, 3))  # all 3-notes subsets
    all_chords = frozenset(
        Chord.from_name(root, name)
        for root, name in itertools.product(config.chromatic_notes, Chord.name_to_intervals)
    )

    rootless_2_rootfull = {chord.rootless: chord for chord in all_chords}

    return (
        n3
        | P.Map(frozenset)
        | P.Map(SpecificChord)
        | P.ValueBy(lambda chord: rootless_2_rootfull.get(chord.abstract))
        | P.FilterValues()
        | P.Append(lambda x: f'{x[1].root.name} {x[1].name}')
        | P.Pipe(tuple)
    )


@functools.cache
def iter_inversions(chord: Chord, octaves):
    notes_iterators = []
    for note in chord.notes:
        notes_iterators.append([SpecificNote(note, octave) for octave in octaves])
    return itertools.product(*notes_iterators) | P.Map(lambda notes: SpecificChord(frozenset(notes), root=chord.root)) | P.Pipe(tuple)


def iter_inversions_chord_progression(progression, octaves):
    inversions = []
    for chord in progression:
        inversions.append(iter_inversions(chord, octaves))
    yield from itertools.product(*inversions)


@functools.cache
def check_all_transitions_not(p, f, *args):
    n = len(p)
    return all(not f(p[i], p[(i + 1) % n], *args) for i in range(n))


def no_double_leading_tone(p, s: Scale):
    '''leading tone here is 7th and 2nd not of a scale, maybe'''
    return all


def unique_roots(progression):
    return len(progression) == len(frozenset(chord.root for chord in progression))


def c_not_c(chords: Iterable[SpecificChord]) -> tuple[list[SpecificChord], list[SpecificChord]]:
    c_chords, not_c_chords = [], []
    for c in chords:
        if c.root.name == 'C':
            c_chords.append(c)
        else:
            not_c_chords.append(c)
    return c_chords, not_c_chords


def notes_are_chord(notes: tuple, scale_chords: frozenset[Chord]):
    abstract = tuple(n.abstract for n in notes)
    abstract_fz = frozenset(abstract)

    for chord in scale_chords:
        if abstract_fz == chord.notes:
            root = chord.root
            break
    else:
        return
    chord = SpecificChord(frozenset(notes), root)
    if chord.notes_ascending[0].abstract != root:
        return
    yield chord


def possible_chords(scale: Scale, note_range: tuple[SpecificNote]) -> tuple[SpecificChord]:
    return (
        note_range
        | P.Filter(lambda note: note.abstract in scale.notes)
        | P.Pipe(lambda it: itertools.combinations(it, 4))  # 4 voice chords
        | P.FlatMap(lambda notes: notes_are_chord(notes, frozenset(chord for chord in scale.triads if chord.name != 'diminished')))
        | P.Filter(checks.no_large_spacing)
        | P.Pipe(tuple)
    )


checks_ = (
    lambda a, b: checks.have_parallel_interval(a, b, 0),
    lambda a, b: checks.have_parallel_interval(a, b, 7),
    lambda a, b: checks.have_hidden_parallel(a, b, 0),
    lambda a, b: checks.have_hidden_parallel(a, b, 7),
    lambda a, b: checks.have_voice_overlap(a, b),
    lambda a, b: checks.have_large_leaps(a, b, 5),
)


def progression_dist(p):
    n = len(p)
    return sum(abs(p[i] - p[(i + 1) % n]) for i in range(n))


@functools.cache
def no_bad_checks(a: SpecificChord, b: SpecificChord):
    return all(not check(a, b) for check in checks_)


def transpose_uniqiue_key(progression: tuple[SpecificChord]):
    origin = progression[0].notes_ascending[0]

    return (
        origin.abstract.i,
        tuple(frozenset(note - origin for note in chord.notes) for chord in progression)
    )


def make_progressions(
    scale: Scale,
    note_range: tuple[SpecificNote],
    n=4,
):
    return (
        sequence_builder(
            n,
            options=possible_chords(scale, note_range),
            curr_prev_constraint=no_bad_checks,
            i_constraints={0: lambda chord: chord.root == scale.root},
            unique_key=lambda chord: chord.root,
        )
        | P.Pipe(lambda it: unique(it, key=transpose_uniqiue_key))
        | P.KeyBy(progression_dist)
        | P.Pipe(lambda x: sorted(x, key=operator.itemgetter(0)))
        | P.Pipe(tuple)
    )


@functools.cache
def all_chords(chord: Chord, note_range, n_notes: int = 3):
    chord_notes = tuple(n for n in note_range if n.abstract in chord.notes)
    return (
        itertools.combinations(chord_notes, n_notes)
        | P.Filter(lambda notes: frozenset(n.abstract for n in notes) == chord.notes)
        | P.Map(lambda notes: SpecificChord(notes, root=chord.root))
        | P.Filter(checks.no_large_spacing)
        | P.Pipe(tuple)
    )


def make_progressions_v2(
    abstract_progression: tuple[Chord],
    n_notes: int = 3,
):
    chord_2_all_chords = tuple(all_chords(chord, config.note_range, n_notes) for chord in abstract_progression)
    return (
        sequence_builder(
            n=len(abstract_progression),
            options=chord_2_all_chords,
            options_separated=True,
            curr_prev_constraint=no_bad_checks,
        )
        | P.Pipe(lambda it: unique(it, key=transpose_uniqiue_key))
        | P.KeyBy(progression_dist)
        | P.Pipe(lambda x: sorted(x, key=operator.itemgetter(0)))
        | P.Pipe(tuple)
    )


def random_progression(s: Scale, n: int = 8, parallel_prob=0.2):
    assert s.kind == 'diatonic'
    parallel_ = parallel(s)
    relative_ = relative(s)
    print('scale', 'parallel', 'relative')
    print(s, parallel_, relative_)
    print('=' * 100)
    chords = []
    chords.append(s.triads[0])

    steps = {'major': (0, 1, 2, 3, 4, 5), 'minor': (0, 2, 3, 4, 5, 6)}[s.name]  # disable diminished

    for _ in range(n - 1):
        step = random.choice(steps)
        s_ = s if random.random() > parallel_prob else parallel_
        c = s_.triads[step]
        print(s_, c)
        chords.append(c)
    return chords


def chord_transitons(
    chord: SpecificChord,
    note_range: tuple[SpecificNote],
    unique_abstract: bool = True,
) -> frozenset[SpecificChord]:
    out = set()
    note_to_i = {note: i for i, note in enumerate(note_range)}

    for note in chord.notes_ascending:
        for add in (-1, 1):
            i = note_to_i[note] + add
            if i == 0 or i == len(note_range):
                continue
            notes = chord.notes - {note} | {note_range[i]}
            if len(notes) != 3:
                continue
            if unique_abstract and len(notes) > len({n.abstract for n in notes}):
                continue
            out.add(SpecificChord(notes))
    return frozenset(out)


def transition_graph(start_chord: SpecificChord, note_range: tuple[SpecificNote]) -> dict[SpecificChord, frozenset[SpecificChord]]:
    graph = collections.defaultdict(set)

    def _graph(chord: SpecificChord):
        if chord in graph:
            return
        childs = chord_transitons(chord, note_range)
        for child in childs:
            graph[chord].add(child)
        for child in childs:
            _graph(child)

    _graph(start_chord)
    graph = dict(graph)
    return graph



import functools
import itertools
from collections import defaultdict
from collections import deque

from musictool import chromatic
from musictool import config
from musictool.chord import Chord
from musictool.note import Note
from musictool.note import SpecificNote
from musictool.noteset import NoteSet
from musictool.piano import Piano
from musictool.util.color import hex_to_rgb
from musictool.util.iteration import iter_scales
from musictool.util.text import cprint


class Scale(NoteSet):
    intervals_to_name = {
        # diatonic
        frozenset({2, 4, 5, 7, 9, 11}): 'major',
        frozenset({2, 3, 5, 7, 9, 10}): 'dorian',
        frozenset({1, 3, 5, 7, 8, 10}): 'phrygian',
        frozenset({2, 4, 6, 7, 9, 11}): 'lydian',
        frozenset({2, 4, 5, 7, 9, 10}): 'mixolydian',
        frozenset({2, 3, 5, 7, 8, 10}): 'minor',
        frozenset({1, 3, 5, 6, 8, 10}): 'locrian',
        # pentatonic
        frozenset({2, 4, 7, 9}): 'p_major',
        frozenset({2, 5, 7, 10}): 'p_dorian',
        frozenset({3, 5, 8, 10}): 'p_phrygian',
        frozenset({2, 5, 7, 9}): 'p_mixolydian',
        frozenset({3, 5, 7, 10}): 'p_minor',
        # sudu
        frozenset({2, 4, 5, 7, 9}): 's_major',
        frozenset({2, 3, 5, 7, 10}): 's_dorian',
        frozenset({1, 3, 5, 8, 10}): 's_phrygian',
        frozenset({2, 4, 7, 9, 11}): 's_lydian',
        frozenset({2, 5, 7, 9, 10}): 's_mixolydian',
        frozenset({3, 5, 7, 8, 10}): 's_minor',
    }
    name_to_intervals = {v: k for k, v in intervals_to_name.items()}

    _cache = {}

    def __new__(
        cls,
        notes: frozenset[str | Note],
        root: str | Note,
    ):
        key = notes, root
        if instance := cls._cache.get(key):
            return instance
        instance = super().__new__(cls)
        instance._long_init(notes, root)
        cls._cache[key] = instance
        return instance

    # def __init__(
    def _long_init(
        self,
        notes: frozenset[str | Note],
        root: str | Note,
    ):
        super().__init__(notes, root)
        self.kind = config.kinds.get(self.name)
        if self.kind == 'diatonic':
            self.triads = self._make_triads()

        self.note_colors = {}
        self.note_scales = {}

        for note, scale in zip(self.notes_ascending, iter_scales(self.kind, start=self.name)):
            self.note_colors[note] = hex_to_rgb(config.scale_colors[scale])
            self.note_scales[note] = scale

        self.html_classes = ('card', self.name)

    def _make_triads(self):
        notes_deque = deque(self.notes_ascending)
        triads = []
        for _ in range(len(notes_deque)):
            chord = Chord(frozenset({notes_deque[0], notes_deque[2], notes_deque[4]}), root=notes_deque[0])
            triads.append(chord)
            notes_deque.rotate(-1)
        return tuple(triads)

    def to_piano_image(self):
        return Piano(scale=self)._repr_svg_()

    def with_html_classes(self, classes: tuple):
        prev = self.html_classes
        self.html_classes = prev + classes
        r = self._repr_html_()
        self.html_classes = prev
        return r

    def _repr_html_(self):
        # <code>bits: {self.bits}</code><br>
        # chords_hover = f"title='{self._chords_text()}'" if self.kind =='diatonic' else ''
        chords_hover = ''
        return f'''
        <div class='{' '.join(self.html_classes)}' {chords_hover}>
        <a href='{self.root.name}'><span class='card_header'><h3>{self.root.name} {self.name}</h3></span></a>
        {self.to_piano_image()}
        </div>
        '''

# flake8: noqa


class ScaleOld:
    _cache = {}

    def __new__(cls, root: str | Note, name: str):
        key = root, name
        if instance := cls._cache.get(key):
            return instance
        instance = super().__new__(cls)
        instance._long_init(root, name)
        cls._cache[key] = instance
        return instance

    def __getnewargs__(self):
        """
        tells pickle.dumps to pickle x in such a way that a pickle.loads
        back from that string will use __new__ with the proper argument.
        """
        return self.root, self.name

    def _long_init(self, root: str | Note, name: str):
        if isinstance(root, str):
            root = Note(root)
        self.root = root
        self.name = name
        self.key = root, name
        self.bits = name_2_bits[name]
        self.notes = tuple(itertools.compress(chromatic.iterate(start_note=root), map(int, self.bits)))

        # self.notes = tuple(itertools.compress(chromatic(root), map(int, self.bits)))
        # print(self.notes)
        self.chromatic_mask = ''.join(note.name if note in self.notes else '_' for note in chromatic.iterate(take_n=12))
        # self.chromatic_bits = ''.join(str(int(note in self.notes)) for note in config.chromatic_notes) # from C (config.chromatic_notes[0])
        # self.chromatic_bits = int(self.bits, base=2)
        self.kind = config.kinds.get(name)
        if self.kind == 'diatonic':
            self.add_chords()

        self.note_colors = {}
        self.note_scales = {}

        for note, scale in zip(self.notes, iter_scales(self.kind, start=self.name)):
            self.note_colors[note] = hex_to_rgb(config.scale_colors[scale])
            self.note_scales[note] = scale

        self.html_classes = ('card', self.name)

    def add_chords(self):
        notes_deque = deque(self.notes)
        chords = []
        for _ in range(len(notes_deque)):
            chord = Chord(frozenset({notes_deque[0], notes_deque[2], notes_deque[4]}), root=notes_deque[0])
            chords.append(chord)
            notes_deque.rotate(-1)
        self.chords = tuple(chords)

    def to_piano_image(self):
        return Piano(scale=self)._repr_svg_()

    def _chords_text(self):
        x = 'chords:\n'
        for i, chord in enumerate(self.chords, start=1):
            x += f'{i} {chord} {chord.name}\n'
        return x

    def with_html_classes(self, classes: tuple):
        prev = self.html_classes
        self.html_classes = prev + classes
        r = self._repr_html_()
        self.html_classes = prev
        return r

    # def __format__(self, format_spec): raise No

    # @functools.cached_property
    def _repr_html_(self):
        # <code>bits: {self.bits}</code><br>
        # chords_hover = f"title='{self._chords_text()}'" if self.kind =='diatonic' else ''
        chords_hover = ''
        return f'''
        <div class='{' '.join(self.html_classes)}' {chords_hover}>
        <a href='{self.root.name}'><span class='card_header'><h3>{self.root.name} {self.name}</h3></span></a>
        {self.to_piano_image()}
        </div>
        '''

    def add_note(self, note: Note | SpecificNote, steps: int):
        if type(note) is Note:
            return self.notes[(self.notes.index(note) + steps) % len(self.notes)]
        elif type(note) is SpecificNote:
            octave, i = divmod(self.notes.index(note.abstract) + steps, len(self.notes))
            return SpecificNote(self.notes[i], note.octave + octave)
        else:
            raise TypeError

    def __eq__(self, other): return self.key == other.key
    def __hash__(self): return hash(self.key)
    def __repr__(self): return f'Scale({self.root} {self.name})'


class ComparedScales:
    '''
    this is compared scale
    local terminology: left sclae is compared to right
    left is kinda parent, right is kinda child
    '''

    def __init__(self, left: Scale, right: Scale):
        self.left = left
        self.right = right
        self.key = left, right
        self.shared_notes = frozenset(left.notes) & frozenset(right.notes)
        self.new_notes = frozenset(right.notes) - frozenset(left.notes)
        self.del_notes = frozenset(left.notes) - frozenset(right.notes)
        if right.kind == 'diatonic':
            self.shared_triads = frozenset(left.triads) & frozenset(right.triads)
        self.html_classes = ('card',)

    def with_html_classes(self, classes: tuple):
        prev = self.html_classes
        self.html_classes = prev + classes
        r = self._repr_html_()
        self.html_classes = prev
        return r

    # def __format__(self, format_spec): raise No

    # @functools.cached_property
    def _repr_html_(self):
        # <code>bits: {self.bits}</code><br>
        # chords_hover = f"title='{self._chords_text()}'" if self.kind =='diatonic' else ''
        chords_hover = ''
        return f'''
        <div class='{' '.join(self.html_classes)}' {chords_hover}>
        <a href='{self.right.root.name}'><span class='card_header'><h3>{self.right.root.name} {self.right.name}</h3></span></a>
        {self.to_piano_image()}
        </div>
        '''

    def to_piano_image(self, as_base64=False):

        return Piano(
            scale=self.right,
            red_notes=self.del_notes, green_notes=self.new_notes, blue_notes=self.shared_notes,
            notes_squares={
                chord.root: (
                    hex_to_rgb(config.chord_colors[chord.name]),
                    config.BLUE_COLOR if chord in self.shared_triads else config.BLACK_COLOR,
                    config.BLUE_COLOR if chord in self.shared_triads else config.BLACK_COLOR,
                    chord.str_chord,
                )
                for chord in self.right.triads
            } if self.right.kind == 'diatonic' else dict(),
        )._repr_svg_()

    def __eq__(self, other): return self.key == other.key
    def __hash__(self): return hash(self.key)
    def __repr__(self): return f'ComparedScale({self.left.root} {self.left.name} | {self.right.root} {self.right.name})'


diatonic = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.diatonic)}
pentatonic = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.pentatonic)}
sudu = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.sudu)}
all_scales = {'diatonic': diatonic, 'pentatonic': pentatonic, 'sudu': sudu}

# circle of fifths clockwise
majors = dict(
    diatonic=tuple(diatonic[note, 'major'] for note in 'CGDAEBfdaebF'),
    pentatonic=tuple(pentatonic[note, 'p_major'] for note in 'CGDAEBfdaebF'),
    sudu=tuple(sudu[note, 's_major'] for note in 'CGDAEBfdaebF'),
)


@functools.cache
def neighbors(left: Scale):
    neighs = defaultdict(list)
    for right in all_scales[left.kind].values():
        # if left == right:
        #     continue
        right = ComparedScales(left, right)
        neighs[len(right.shared_notes)].append(right)
    return neighs


def print_neighbors(s: Scale):
    neighs = neighbors(s)
    for n_intersect in sorted(neighs.keys(), reverse=True):
        for n in neighs[n_intersect]:
            if n.name != 'major': continue
            print(repr(n).ljust(32), '|', end=' ')
            for note in n.chromatic_mask:
                if note in s.chromatic_mask: print(cprint(note, color='BLUE'), end='')
                else: print(note, end='')
            print(' |', end=' ')
            for chord in n.triads:
                if chord in n.shared_triads: print(cprint(chord, color='BLUE'), end=' ')
                else: print(chord, end=' ')
            print()
        print('=' * 100)

# warm up cache
# for scale in tqdm.tqdm(tuple(itertools.chain(all_scales['diatonic'].values(), all_scales['pentatonic'].values()))):
#     _ = scale.to_piano_image(as_base64=True)
#     for neighbor in itertools.chain.from_iterable(neighbors(scale).values()):
#         _ = neighbor.to_piano_image(as_base64=True)


def parallel(s: Scale):
    """same root, convert major to minor and vice versa"""
    return Scale.from_name(s.root, {'major': 'minor', 'minor': 'major'}[s.name])


def relative(s: Scale):
    """same set of notes, changes root, convert major to minor and vice versa"""
    relative_name = {'major': 'minor', 'minor': 'major'}[s.name]
    for note, name in s.note_scales.items():
        if name == relative_name:
            return Scale.from_name(note, name)



from collections import defaultdict

chromatic_notes = 'CdDeEFfGaAbB'  # todo make variable here, delete from config, reimport everywhere, maybe circular imports
note_i = {note: i for i, note in enumerate(chromatic_notes)}
is_black = {note: bool(int(x)) for note, x in zip(chromatic_notes, '010100101010')}

neighsbors_min_shared = {'diatonic': 0, 'pentatonic': 0}


diatonic = 'major', 'dorian', 'phrygian', 'lydian', 'mixolydian', 'minor', 'locrian'
pentatonic = 'p_major', 'p_dorian', 'p_phrygian', 'p_mixolydian', 'p_minor'
sudu = 's_major', 's_dorian', 's_phrygian', 's_lydian', 's_mixolydian', 's_minor'
kinds = {k: 'diatonic' for k in diatonic} | {k: 'pentatonic' for k in pentatonic} | {k: 'sudu' for k in sudu}


# if change: also change in static/main.css
scale_colors = dict(
    major='FFFFFF',
    dorian='54E346',
    phrygian='00FFCC',
    lydian='68A6FC',
    mixolydian='FFF47D',
    minor='D83A56',
    locrian='B980F0',
)
scale_colors['p_major'] = scale_colors['major']
scale_colors['p_dorian'] = scale_colors['dorian']
scale_colors['p_phrygian'] = scale_colors['phrygian']
scale_colors['p_mixolydian'] = scale_colors['mixolydian']
scale_colors['p_minor'] = scale_colors['minor']

scale_colors['s_major'] = scale_colors['major']
scale_colors['s_dorian'] = scale_colors['dorian']
scale_colors['s_phrygian'] = scale_colors['phrygian']
scale_colors['s_lydian'] = scale_colors['lydian']
scale_colors['s_mixolydian'] = scale_colors['mixolydian']
scale_colors['s_minor'] = scale_colors['minor']

WHITE_COLOR = (0xaa,) * 3
BLACK_COLOR = (0x50,) * 3
WHITE = 0xFF, 0xFF, 0xFF
BLACK = 0, 0, 0
RED_COLOR = 0xff, 0, 0
GREEN_COLOR = 0, 0xff, 0
BLUE_COLOR = 0, 0, 0xff

chord_colors = {
    'minor': scale_colors['minor'],
    'major': scale_colors['major'],
    'diminished': scale_colors['locrian'],
}

default_octave = 5
# DEFAULT_TUNING = 440  # default A hz tuning
DEFAULT_TUNING = 500
RANDOM_TUNING_RANGE = 420, 510
tuning = DEFAULT_TUNING


# piano_img_size = 14 * 60, 280
piano_img_size = 14 * 18, 85
beats_per_minute = 120
beats_per_second = beats_per_minute / 60
beats_per_bar = 4
bar_seconds = beats_per_bar / beats_per_second


# daw
sample_rate = 44100  # samples per second
midi_folder = 'static/midi/'
# midi_file = 'weird.mid'
# midi_file = 'overlap.mid'
# midi_file = 'dots.mid'
# midi_file = 'halfbar.mid'
# midi_file = 'halfbar-and-short.mid'
# midi_file = 'bassline.mid'
# midi_file = 'drumloop.mid'
# midi_file = '4-4-8.mid'
# chunk_size = 1024 * 32
chunk_size = 1024
# chunk_size = 1024 * 2
# chunk_size = 1024 * 128
# chunk_size = 1024 * 4
chunk_seconds = chunk_size / sample_rate
wav_output_file = 'out.wav'

# samples
kick = 'static/samples/kick-909.wav'
hat = 'static/samples/open-hat-909.wav'
clap = 'static/samples/clap-909.wav'


# streaming
audio_pipe = 'audio.fifo'
video_pipe = 'video.fifo'

# fps = 24
# fps = 30
# fps = 48
# fps = 50
# fps = 44
# fps = 40
# fps = 30
# fps = 55
fps = 60

# frame_width, frame_height = 2560, 1440
frame_width, frame_height = 1920, 1080  # 1080p, recommended bitrate 4.5M
# frame_width, frame_height = 1280, 720  # 720p
# frame_width, frame_height = 854, 480  # 480p
# frame_width, frame_height = 640, 360 # 360p
# frame_width, frame_height = 426, 240  # 240p
# video_bitrate = '500k'
# video_bitrate = '3000k'
# video_bitrate = '3m'
audio_bitrate = '128k'
# video_bitrate = '12M'
# video_bitrate = '24M'
# keyframe_seconds = 0.05  # drastically changes bitrate
keyframe_seconds = 0.25  # drastically changes bitrate
# keyframe_seconds = 0.5  # drastically changes bitrate

gop = int(keyframe_seconds * fps)
draw_threads = 1
# assert draw_threads == 1

bars_per_screen = 4
screen_seconds = bars_per_screen * bar_seconds

chord_px = frame_height / bars_per_screen
# pxps = frame_height // screen_seconds  # pixels per second
pxps = frame_height / screen_seconds  # pixels per second

# video_queue_item_size = 20
video_queue_item_size = 1
assert video_queue_item_size < fps


# OUTPUT_VIDEO = '/tmp/output.flv'
# OUTPUT_VIDEO = '/dev/null'
OUTPUT_VIDEO = None

log_path = 'logs/log.jsonl'


note_range = None
messages = []

progressions = None
progressions_queue = None


progressions_search_cache = defaultdict(list)

ui_thread = None


# midi explorer ui

# MIDI_UI_FILE = 'static/midi/vespers-04.mid'
MIDI_UI_FILE = 'static/midi/vivaldi-winter.mid'
